import functools
import http.cookies
import secrets
from re import Pattern
from typing import Dict, List, Optional, Set, cast

from itsdangerous import BadSignature
from itsdangerous.url_safe import URLSafeSerializer
from starlette.datastructures import URL, MutableHeaders
from starlette.requests import Request
from starlette.responses import PlainTextResponse, Response
from starlette.types import ASGIApp, Message, Receive, Scope, Send


class CSRFMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        secret: str,
        sensitive_cookies: Optional[Set[str]] = None,
        cookie_name: str = "csrftoken",
        cookie_path: str = "/",
        cookie_domain: Optional[str] = None,
        cookie_secure: bool = False,
        cookie_httponly: bool = False,
        cookie_samesite: str = "lax",
        header_name: str = "x-csrftoken",
        safe_methods: Set[str] = {"GET", "HEAD", "OPTIONS", "TRACE"},
        exempt_urls: Optional[List[Pattern]] = None,
    ) -> None:
        self.app = app
        self.serializer = URLSafeSerializer(secret, "csrftoken")
        self.secret = secret
        self.sensitive_cookies = sensitive_cookies
        self.cookie_name = cookie_name
        self.cookie_path = cookie_path
        self.cookie_domain = cookie_domain
        self.cookie_secure = cookie_secure
        self.cookie_httponly = cookie_httponly
        self.cookie_samesite = cookie_samesite
        self.header_name = header_name
        self.safe_methods = safe_methods
        self.exempt_urls = exempt_urls

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):  # pragma: no cover
            await self.app(scope, receive, send)
            return

        request = Request(scope)
        csrf_cookie = request.cookies.get(self.cookie_name)

        if (
            request.method not in self.safe_methods
            and self._has_sensitive_cookies(request.cookies)
            and not self._url_is_exempt(request.url)
        ):
            submitted_csrf_token = await self._get_submitted_csrf_token(request)
            if (
                not csrf_cookie
                or not submitted_csrf_token
                or not self._csrf_tokens_match(csrf_cookie, submitted_csrf_token)
            ):
                response = self._get_error_response(request)
                await response(scope, receive, send)
                return

        send = functools.partial(self.send, send=send, scope=scope)
        await self.app(scope, receive, send)

    async def send(self, message: Message, send: Send, scope: Scope) -> None:
        request = Request(scope)
        csrf_cookie = request.cookies.get(self.cookie_name)

        if csrf_cookie is None:
            message.setdefault("headers", [])
            headers = MutableHeaders(scope=message)

            cookie: http.cookies.BaseCookie = http.cookies.SimpleCookie()
            cookie_name = self.cookie_name
            cookie[cookie_name] = self._generate_csrf_token()
            cookie[cookie_name]["path"] = self.cookie_path
            cookie[cookie_name]["secure"] = self.cookie_secure
            cookie[cookie_name]["httponly"] = self.cookie_httponly
            cookie[cookie_name]["samesite"] = self.cookie_samesite
            if self.cookie_domain is not None:
                cookie[cookie_name]["domain"] = self.cookie_domain  # pragma: no cover
            headers.append("set-cookie", cookie.output(header="").strip())

        await send(message)

    def _has_sensitive_cookies(self, cookies: Dict[str, str]) -> bool:
        if not self.sensitive_cookies:
            return True
        for sensitive_cookie in self.sensitive_cookies:
            if sensitive_cookie in cookies:
                return True
        return False

    def _url_is_exempt(self, url: URL) -> bool:
        if not self.exempt_urls:
            return False
        for exempt_url in self.exempt_urls:
            if exempt_url.match(url.path):
                return True
        return False

    async def _get_submitted_csrf_token(self, request: Request) -> Optional[str]:
        return request.headers.get(self.header_name)

    def _generate_csrf_token(self) -> str:
        return cast(str, self.serializer.dumps(secrets.token_urlsafe(128)))

    def _csrf_tokens_match(self, token1: str, token2: str) -> bool:
        try:
            decoded1: str = self.serializer.loads(token1)
            decoded2: str = self.serializer.loads(token2)
            return secrets.compare_digest(decoded1, decoded2)
        except BadSignature:
            return False

    def _get_error_response(self, request: Request) -> Response:
        return PlainTextResponse(
            content="CSRF token verification failed", status_code=403
        )



import asyncio
import re
from typing import Callable

import httpx
import pytest
from asgi_lifespan import LifespanManager
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.routing import Route

from starlette_csrf import CSRFMiddleware


@pytest.fixture(scope="session")
def event_loop():
    """Force the pytest-asyncio loop to be the main one."""
    loop = asyncio.get_event_loop()
    yield loop


@pytest.fixture(scope="session")
def all_sensitive_csrfmiddleware() -> Middleware:
    return Middleware(CSRFMiddleware, secret="SECRET")


@pytest.fixture(scope="session")
def some_sensitive_csrfmiddleware() -> Middleware:
    return Middleware(CSRFMiddleware, secret="SECRET", sensitive_cookies={"sensitive"})


@pytest.fixture(scope="session")
def some_exempt_csrfmiddleware() -> Middleware:
    return Middleware(
        CSRFMiddleware, secret="SECRET", exempt_urls=[re.compile(r"/exempt")]
    )


@pytest.fixture
def app_generator() -> Callable[[Middleware], Starlette]:
    def _app_generator(middleware: Middleware) -> Starlette:
        async def get(request: Request):
            return JSONResponse({"hello": "world"})

        async def post(request: Request):
            json = await request.json()
            return JSONResponse(json)

        app = Starlette(
            debug=True,
            routes=[
                Route("/get", get, methods=["GET"]),
                Route("/post", post, methods=["POST"]),
                Route("/exempt", post, methods=["POST"]),
            ],
            middleware=[middleware],
        )

        return app

    return _app_generator


@pytest.fixture
async def test_client_all_sensitive(app_generator, all_sensitive_csrfmiddleware):
    app = app_generator(all_sensitive_csrfmiddleware)
    async with LifespanManager(app):
        async with httpx.AsyncClient(app=app, base_url="http://app.io") as test_client:
            yield test_client


@pytest.fixture
async def test_client_some_sensitive(app_generator, some_sensitive_csrfmiddleware):
    app = app_generator(some_sensitive_csrfmiddleware)
    async with LifespanManager(app):
        async with httpx.AsyncClient(app=app, base_url="http://app.io") as test_client:
            yield test_client


@pytest.fixture
async def test_client_some_exempt(app_generator, some_exempt_csrfmiddleware):
    app = app_generator(some_exempt_csrfmiddleware)
    async with LifespanManager(app):
        async with httpx.AsyncClient(app=app, base_url="http://app.io") as test_client:
            yield test_client



from typing import Dict

import httpx
import pytest
from starlette import status


@pytest.mark.asyncio
async def test_get(test_client_all_sensitive: httpx.AsyncClient):
    response = await test_client_all_sensitive.get("/get")

    assert response.status_code == status.HTTP_200_OK

    assert "csrftoken" in response.cookies

    set_cookie_header = response.headers["set-cookie"]
    assert "Path=/;" in set_cookie_header
    assert "HttpOnly" not in set_cookie_header
    assert "Secure" not in set_cookie_header


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "cookies,headers",
    [
        ({}, {}),
        ({}, {"x-csrftoken": "aaa"}),
        ({"csrftoken": "aaa"}, {}),
        ({"csrftoken": "aaa"}, {"x-csrftoken": "aaa"}),
    ],
)
async def test_post_invalid_csrf(
    test_client_all_sensitive: httpx.AsyncClient,
    cookies: Dict[str, str],
    headers: Dict[str, str],
):
    response = await test_client_all_sensitive.post(
        "/post", cookies=cookies, headers=headers
    )

    assert response.status_code == status.HTTP_403_FORBIDDEN


@pytest.mark.asyncio
async def test_valid_csrf(test_client_all_sensitive: httpx.AsyncClient):
    response_get = await test_client_all_sensitive.get("/get")
    csrf_cookie = response_get.cookies["csrftoken"]

    response_post = await test_client_all_sensitive.post(
        "/post",
        headers={"x-csrftoken": csrf_cookie},
        json={"hello": "world"},
    )

    assert response_post.status_code == status.HTTP_200_OK


@pytest.mark.asyncio
async def test_some_sensitive_csrf(test_client_some_sensitive: httpx.AsyncClient):
    response_get = await test_client_some_sensitive.get("/get")
    csrf_cookie = response_get.cookies["csrftoken"]

    response_post_not_sensitive = await test_client_some_sensitive.post(
        "/post",
        cookies={"foo": "bar"},
        json={"hello": "world"},
    )

    assert response_post_not_sensitive.status_code == status.HTTP_200_OK

    response_post_sensitive_no_csrf_token = await test_client_some_sensitive.post(
        "/post",
        cookies={"sensitive": "bar"},
        json={"hello": "world"},
    )

    assert (
        response_post_sensitive_no_csrf_token.status_code == status.HTTP_403_FORBIDDEN
    )

    response_post_sensitive_csrf_token = await test_client_some_sensitive.post(
        "/post",
        cookies={"sensitive": "bar"},
        headers={"x-csrftoken": csrf_cookie},
        json={"hello": "world"},
    )

    assert response_post_sensitive_csrf_token.status_code == status.HTTP_200_OK


@pytest.mark.asyncio
async def test_some_exempt_csrf(test_client_some_exempt: httpx.AsyncClient):
    response_post_exempt = await test_client_some_exempt.post(
        "/exempt",
        cookies={"foo": "bar"},
        json={"hello": "world"},
    )

    assert response_post_exempt.status_code == status.HTTP_200_OK

    response_post_not_exempt = await test_client_some_exempt.post(
        "/post",
        cookies={"sensitive": "bar"},
        json={"hello": "world"},
    )

    assert response_post_not_exempt.status_code == status.HTTP_403_FORBIDDEN



# -*- coding: utf-8 -*-
from setuptools import setup

setup(
    name='starlette-csrf',
    version='1.4.5',
    description='Starlette middleware implementing Double Submit Cookie technique to mitigate CSRF',
    long_description='# Starlette CSRF Middleware\n\nStarlette middleware implementing Double Submit Cookie technique to mitigate CSRF.\n\n[![build](https://github.com/frankie567/starlette-csrf/workflows/Build/badge.svg)](https://github.com/frankie567/starlette-csrf/actions)\n[![codecov](https://codecov.io/gh/frankie567/starlette-csrf/branch/main/graph/badge.svg?token=fL49kIvrj6)](https://codecov.io/gh/frankie567/starlette-csrf)\n[![PyPI version](https://badge.fury.io/py/starlette-csrf.svg)](https://badge.fury.io/py/starlette-csrf)\n[![Downloads](https://pepy.tech/badge/starlette-csrf)](https://pepy.tech/project/starlette-csrf)\n\n<p align="center">\n<a href="https://github.com/sponsors/frankie567"><img src="https://md-btn.deta.dev/button.svg?text=Buy%20me%20a%20coffee%20%E2%98%95%EF%B8%8F&bg=ef4444&w=200&h=50"></a>\n</p>\n\n## How it works?\n\n1. The user makes a first request with a method considered safe (by default `GET`, `HEAD`, `OPTIONS`, `TRACE`).\n2. It receives in response a cookie (named by default `csrftoken`) which contains a secret value.\n3. When the user wants to make an unsafe request, the server expects them to send the same secret value in a header (named by default `x-csrftoken`).\n4. The middleware will then compare the secret value provided in the cookie and the header.\n   * If they match, the request is processed.\n   * Otherwise, a `403 Forbidden` error response is given.\n\nThis mechanism is necessary if you rely on cookie authentication in a browser. You can have more information about CSRF and Double Submit Cookie in the [OWASP Cheat Sheet Series](https://cheatsheetseries.owasp.org/cheatsheets/Cross-Site_Request_Forgery_Prevention_Cheat_Sheet.html).\n\n## Installation\n\n```bash\npip install starlette-csrf\n```\n\n## Usage with Starlette\n\n```py\nfrom starlette.applications import Starlette\nfrom starlette.middleware import Middleware\nfrom starlette_csrf import CSRFMiddleware\n\nroutes = ...\n\nmiddleware = [\n    Middleware(CSRFMiddleware, secret="__CHANGE_ME__")\n]\n\napp = Starlette(routes=routes, middleware=middleware)\n```\n\n## Usage with FastAPI\n\n```py\nfrom fastapi import FastAPI\nfrom starlette_csrf import CSRFMiddleware\n\napp = FastAPI()\n\napp.add_middleware(CSRFMiddleware, secret="__CHANGE_ME__")\n```\n\n## Arguments\n\n* `secret` (`str`): Secret to sign the CSRF token value. **Be sure to choose a strong passphrase and keep it SECRET**.\n* `sensitive_cookies` (`Set[str]` - `None`): Set of cookie names that should trigger the CSRF check if they are present in the request. Useful if you have other authentication methods that don\'t rely on cookies and don\'t need CSRF enforcement. If this parameter is `None`, the default, CSRF is **always** enforced.\n* `cookie_name` (`str` - `csrftoken`): Name of the cookie.\n* `cookie_path` `str` - `/`): Cookie path.\n* `cookie_domain` (`Optional[str]` - `None`): Cookie domain. If your frontend and API lives in different sub-domains, be sure to set this argument with your root domain to allow your frontend sub-domain to read the cookie on the JavaScript side.\n* `cookie_secure` (`bool` - `False`): Whether to only send the cookie to the server via SSL request.\n* `cookie_samesite` (`str` - `lax`): Samesite strategy of the cookie.\n* `header_name` (`str` - `x-csrftoken`): Name of the header where you should set the CSRF token.\n* `safe_methods` (`Set[str]` - `{"GET", "HEAD", "OPTIONS", "TRACE"}`): HTTP methods considered safe which don\'t need CSRF protection.\n* `exempt_urls` (`Optional[List[re.Pattern]]` - `None`): List of URL regexes that the CSRF check should be skipped on. Useful if you have any APIs that you know do not need CSRF protection.\n\n## Customize error response\n\nBy default, a plain text response with the status code 403 is returned when the CSRF verification is failing. You can customize it by overloading the middleware class and implementing the `_get_error_response` method. It accepts in argument the original `Request` object and expects a `Response`. For example:\n\n```py\nfrom starlette.requests import Request\nfrom starlette.responses import JSONResponse, Response\nfrom starlette_csrf import CSRFMiddleware\n\nclass CustomResponseCSRFMiddleware(CSRFMiddleware):\n    def _get_error_response(self, request: Request) -> Response:\n        return JSONResponse(\n            content={"code": "CSRF_ERROR"}, status_code=403\n        )\n```\n\n## Development\n\n### Setup environment\n\nWe use [Hatch](https://hatch.pypa.io/latest/install/) to manage the development environment and production build. Ensure it\'s installed on your system.\n\n### Run unit tests\n\nYou can run all the tests with:\n\n```bash\nhatch run test\n```\n\n### Format the code\n\nExecute the following command to apply `isort` and `black` formatting:\n\n```bash\nhatch run lint\n```\n\n## License\n\nThis project is licensed under the terms of the MIT license.\n',
    author_email='François Voron <fvoron@gmail.com>',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet :: WWW/HTTP :: Session',
    ],
    install_requires=[
        'itsdangerous<3.0.0,>=2.0.1',
        'starlette>=0.14.2',
    ],
    packages=[
        'starlette_csrf',
        'tests',
    ],
)
